/*
  ikkeiplexing.cpp - Using MsTimer2() with 1ms resolution
  by ikkei <ikkei@zeus.eonet.ne.jp> http://blog.goo.ne.jp/jh3kxm
  
  
  Original:
  Charlieplexing.cpp
  Alex Wenger <a.wenger@gmx.de> http://arduinobuch.wordpress.com/
  Matt Mets <mahto@cibomahto.com> http://cibomahto.com/
  
  Timer init code from MsTimer2 - Javier Valencia <javiervalencia80@gmail.com>
  Misc functions from Benjamin Sonnatg <benjamin@sonntag.fr>
  
  History:
    2009-12-30 - V0.0 wrote the first version at 26C3/Berlin
    2010-01-01 - V0.1 adding misc utility functions 
      (Clear, Vertical,  Horizontal) comment are Doxygen complaints now
    2010-05-27 - V0.2 add double-buffer mode
    
    2010-11-24
      Modified for ikkeiplexing shield (direct two 8x8 dot matrix LED shield)
      With brightness equalization
	  by ikkei <ikkei@zeus.eonet.ne.jp> http://blog.goo.ne.jp/jh3kxm
	2012-12-23
	  Modified for Titchaino with switch
	  by ikkei
	2015-03-16
	  Modified for Titchaino_uni with switch
      by ikkei
	2018-07-13
	  Rename ikkeiplexing.cpp from Charlieplexing.cpp
      by ikkei

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <arduino.h> // for Arduino 1.0
#include <avr/interrupt.h>
#include "ikkeiplexing.h"
#include <MsTimer2.h>


const byte maptable[][16] PROGMEM = {
	{  1,  4, 15,  7, 16,  8, 11, 12,  6, 10, 17,  2, 14,  3,  9,  5},	// IKKEIPLEXING		for ikkeiplexing shield
	{ 17, 13,  9, 11,  3, 15,  4,  7, 12,  5,  6, 16,  8, 10, 14,  2},	// TITCHAINO		for Titchaino
	{ 17, 13,  9, 11,  3, 15,  4,  7, 12,  5,  6, 16,  8, 10, 14,  2},	// TITCHAINO_UNI	for Titchaino uni
	{ 17, 12, 13, 14,  5,  7,  3,  2,  9,  4,  8,  6, 16, 15, 11, 10},	// ONBOARD			for on board
	{  9,  4, 10,  6, 17, 11, 16, 13,  5, 15, 14,  8, 12,  7,  3,  2},	// DOTSDUINO		for Dotsduino
	{  2,  3,  4, 13, 14, 15, 16, 17, 12, 11, 10,  9,  5,  6,  7,  8},	// DOTSDUINO2D		for Dotsduino 2
	{ 15,  2, 13, 18,  5, 12,  6,  9, 19,  7,  8, 16, 11, 17,  3,  4},	// SQUARE8			for 8 Square
};


/* -----------------------------------------------------------------  */
/** Table for the LED multiplexing cycles, containing 16 cycles made out of three bytes
 */
byte leds[2][48];

/// Determines whether the display is in single or double buffer mode
byte displayMode;

/// Flag indicating that the display page should be flipped as soon as the
/// current frame is displayed
boolean videoFlipPage;

/// Pointer to the buffer that is currently being displayed
byte* displayBuffer;

/// Pointer to the buffer that should currently be drawn to
byte* workBuffer;

//boolian frame = false; //frame, 10ms, 100ms


static int sw1;
static int sw2;
static int sw3;
static int sw4;
static int sw5;
static int sw6;
static int sw7;
static int sw8;
static int sw9;
static int sw10;

static byte assign;


static byte in_d = 0;
static byte in_b = 0;
static byte in_c = 0;
static byte out_d = 0;
static byte out_b = 0;
static byte out_c = 0;
static byte sw1234;
static byte swdata;
static byte period;
static byte current_sw;
static byte previous_sw;
static byte scan;// 16 Cycles of Matrix
static byte msec;

//###
#define MID_LED 5
#define MAX_LED 8
#define FRAME 0x01
#define TENMS 0x02

//### brightness equalization parameter
const int dd[MAX_LED] PROGMEM= {1, 180, 270, 370, 490, 650, 800, 950 };	//### duty ratio

/* -----------------------------------------------------------------  */
/** Table for LED Position in leds[] ram table 
 */

	
byte *ledMap;

// all output Hi-Z
void alloff( void ){
	DDRD &= ~out_d; // line 76543210
	PORTD &= ~out_d;
	DDRB &= ~out_b; // line xxDCBA98
	PORTB &= ~out_b;
	DDRC &= ~out_c; // line xxJIHGFE
	PORTC &= ~out_c;
//	#ifndef titchaino_uni
//	DDRC &= ~0x30; // A4,A5
//	#endif
}

/* -----------------------------------------------------------------  */
/** The Interrupt code goes here !  
 */
void scanLED(){
	alloff(); // all output high-Z (input)

	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD |= in_d;	// pull up register of input port
		PORTB |= in_b;	// pull up register of input port
		PORTC |= in_c;	// pull up register of input port
	}

	byte n = pgm_read_byte(&ledMap[ scan ]); // scan column output LOW for cathode
	//byte n = ledMap[ scan ];
	if (n < 8) {
		DDRD  = _BV(n);
	} else if (n < 14 ){
		DDRB  = _BV(n-8);
	} else {
		DDRC  = _BV(n-14);
		//DDRC  = _BV(n-14) | 0x30; // A4,A5
	} 

	//### count ON LED numbers
	byte on_count = 0;
	for (int j=0; j<8; j++){
		if (displayBuffer[scan*3] & 1<<j ){
		    on_count++;
		}
	}
	for (int j=0; j<6; j++){
		if (displayBuffer[scan*3+1] & 1<<j){
		    on_count++;
		}
		if (displayBuffer[scan*3+2] & 1<<j){
		    on_count++;
		}
	}

	if (scan==8){ // get SW1,SW2,SW3,SW4
		sw1234 = (digitalRead(sw1)*8 + digitalRead(sw2)*4 + digitalRead(sw3)*2 + digitalRead(sw4)) ^ 0x0F;
	} else if (scan==9){ // get SW5,SW6,SW7,SW8
		previous_sw = current_sw;
		if (assign == IKKEIPLEXING){
			current_sw = (sw1234 << 4) | ((digitalRead(sw5)*8 + digitalRead(sw6)*4 + digitalRead(sw7)*2 + digitalRead(sw8)) ^ 0x0F);
		} else {
			current_sw = ((digitalRead(sw9)<<1 | digitalRead(sw10)) ^ 0x03) << 6;
		}
		if (previous_sw == current_sw){
			swdata = current_sw;
		}
	} 
	
	// swdata: SW1,SW2,SW3,SW4,SW5,SW6,SW7,SW8
	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD &= ~in_d;	// pull up register of input port
		PORTB &= ~in_b;	// pull up register of input port
		PORTC &= ~in_c;	// pull up register of input port
	}

	//### equalization pre-delay
	if ( (on_count >= MID_LED) && (on_count < MAX_LED) ){
		delayMicroseconds( 1000 - pgm_read_word_near(&dd[on_count]) );
	}

	// output row data
	DDRD  |= displayBuffer[scan*3];
	PORTD  = displayBuffer[scan*3];
	DDRB  |= displayBuffer[scan*3+1];
	PORTB  = displayBuffer[scan*3+1];
	DDRC  |= displayBuffer[scan*3+2];
	PORTC  = (PORTC & 0x30) | displayBuffer[scan*3+2];

	//### equalization post-delay 
	if ( on_count < MID_LED ){
		delayMicroseconds( pgm_read_word_near(&dd[on_count]) );
		alloff();
	}

	period = 0;
	msec++;
	if (msec >= 10){
		msec = 0;
		period |= TENMS;
	}
	
	scan++;
	if (scan >= 16){
		scan = 0;
		period |= FRAME;

		// If the page should be flipped, do it here.
		if (videoFlipPage && displayMode == DOUBLE_BUFFER){
		    // TODO: is this an atomic operation?
		    videoFlipPage = false;

		    byte* temp = displayBuffer;
		    displayBuffer = workBuffer;
		    workBuffer = temp;
		}
	}
}


/* -----------------------------------------------------------------  */
/** Constructor : Initialize the interrupt code. 
 * should be called in setup();
 */
void ikkeiplexing::Init(byte a, byte mode = SINGLE_BUFFER){
	assign = a;
	ledMap = maptable[assign];
	
	sw5 = sw1 = pgm_read_byte(&ledMap[12]);
	sw6 = sw2 = pgm_read_byte(&ledMap[13]);
	sw7 = sw3 = pgm_read_byte(&ledMap[14]);
	sw8 = sw4 = pgm_read_byte(&ledMap[15]);

	
	if (assign == TITCHAINO_UNI){
		sw9  = 0;
		sw10 = 1;
	} else if (assign == SQUARE8){
		sw9  = 14; // A0
		sw10 = 14; // A0
	} else {
		sw9  = 18; // A4
		sw10 = 19; // A5
	}


	//### MsTimer2 setting
	MsTimer2::set(1,scanLED);
	MsTimer2::start();
	
	
	//### make using port bit pattern
	for (int j=0; j<16; j++){
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			out_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    out_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    out_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}
	
	//### make using input port bit pattern
	for (int j=12; j<16; j++){  // LA5,LA6,LA7,LA8
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			in_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    in_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    in_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}

  	
	if (assign == TITCHAINO_UNI){
		in_d |= 0x03; 						// D0,D1 pullup
	} else if (assign != IKKEIPLEXING){
		in_c |= 0x30; 						// A4,A5 pullup
	} else if (assign == SQUARE8){
		in_c |= 0x01; 						// A0 pullup
	}

	// Record whether we are in single or double buffer mode
	displayMode = mode;

	// Point the display buffer to the first physical buffer
	displayBuffer = leds[0];

	// If we are in single buffered mode, point the work buffer
	// at the same physical buffer as the display buffer.  Otherwise,
	// point it at the second physical buffer.
	if( displayMode == SINGLE_BUFFER ) {
	    workBuffer = displayBuffer;
	} else {
	    workBuffer = leds[1];
	}

	scan = 0;
	msec = 0;

	// Clear the buffer and display it
	ikkeiplexing::Clear(0);
	ikkeiplexing::Flip(false);
}


/* -----------------------------------------------------------------  */
/** Constructor : Initialize the interrupt code. 
 * should be called in setup();
 */
/*
void Charlieplexing::Init(byte mode)
//void iLedSign::Init(byte mode)
{
	//### MsTimer2 setting
	MsTimer2::set(1,scanLED);
	MsTimer2::start();

	//### make using port bit pattern
	for (int j=0; j<16; j++){
		//byte k = ledMap[ j ];
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			out_d |= _BV( k      );	// line 76543210
		} else if ( k <14 ){
		    out_b |= _BV( k -  8 ); // line xxDCBA98
		} else {
		    out_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}
	
	//### make using input port bit pattern
	for (int j=12; j<16; j++){  // LA5,LA6,LA7,LA8
		//byte k = ledMap[ j ];
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			in_d  |= _BV( k      );	// line 76543210
		} else if ( k <14 ){
		    in_b  |= _BV( k -  8 );	// line xxDCBA98
		} else {
		    in_c  |= _BV( k - 14 );	// line xxJIHGFE
		}
	}
	if (assign == SQUARE8){
		in_c |= 0x01; // A0 pullup
	} else if (assign == TITCHAINO_UNI){
		in_d |= 0x03; // D0,D1 pullup
	} else if (assign != IKKEIPLEXING){
		in_c |= 0x30; // A4,A5 pullup
	}

	// Record whether we are in single or double buffer mode
	displayMode = mode;

	// Point the display buffer to the first physical buffer
	displayBuffer = leds[0];

	// If we are in single buffered mode, point the work buffer
	// at the same physical buffer as the display buffer.  Otherwise,
	// point it at the second physical buffer.
	if( displayMode == SINGLE_BUFFER ) {
	    workBuffer = displayBuffer;
	}
	else {
	    workBuffer = leds[1];
	}

	scan = 0;
	msec = 0;

	// Clear the buffer and display it
	Charlieplexing::Clear(0);
	Charlieplexing::Flip(false);
}
*/


/* -----------------------------------------------------------------  */
/* * Clear the screen completely
 * @param blocking if true : wait for flip before returning, if false :
 *                 return immediately.
 */
void ikkeiplexing::Flip(bool blocking){
    if (displayMode == DOUBLE_BUFFER){
        // Just set the flip flag, the buffer will flip between redraws
        videoFlipPage = true;

        // If we are blocking, sit here until the page flips.
        while (blocking && videoFlipPage) {
            delay(1);
        }
    }
}


/* -----------------------------------------------------------------  */
/** Clear the screen completely
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Clear(int set){
    for(int x=0;x<16;x++){
        for(int y=0;y<8;y++){
            Set(x,y,set);
        }
    }
}

/* -----------------------------------------------------------------  */
/** Shift left the screen completely
 */
void ikkeiplexing::ShiftLeft(){
    for(int x=1;x<16;x++){
        for(int y=0;y<8;y++){
            Set(x-1,y,Get(x,y));
        }
    }
}

/* -----------------------------------------------------------------  */
/** Shift right the screen completely
 */
void ikkeiplexing::ShiftRight(){
    for(int x=1;x<16;x++){
        for(int y=0;y<8;y++){
            Set(16-x,y,Get(15-x,y));
        }
    }
}

/* -----------------------------------------------------------------  */
/** Clear an horizontal line completely
 * @param y is the y coordinate of the line to clear/light [0-8]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Horizontal(int y, int set){
    for(int x=0;x<16;x++){
        Set(x,y,set);
    }
}


/* -----------------------------------------------------------------  */
/** Clear a vertical line completely
 * @param x is the x coordinate of the line to clear/light [0-13]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Vertical(int x, int set){
    for(int y=0;y<8;y++){
        Set(x,y,set);
    }
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::Vwrite(int x, byte data){
    for(int y=0; y<8; y++){
    	if( data & (1<<y)){
    		Set(x, y, 1);
      	} else {
			Set(x, y, 0);
      	}
    }
}


/* -----------------------------------------------------------------  */
void ikkeiplexing::write(byte y, const uint16_t values[], byte size){
	for(byte i=0;i<size;i++){
//		_buffer[(y++) & 0x07] = values[i];
		for(byte j=0;j<16;j++){
			if ((y+i) < 8){
				Set(j  , ((y+i) & 0x07), bitRead( values[i],15-j));
			} else {
				Set(j+8, ((y+i) & 0x07), bitRead( values[i],15-j));
			}
		}
	}
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::write90(byte y, const uint16_t values[], byte size){
	for(byte i=0;i<size;i++){
//		_buffer[(y++) & 0x07] = values[i];
		for(byte j=0;j<16;j++){
			Set(((y+i) & 0x0F), j,   bitRead( values[i],15-j));
		}
	}
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::Set(byte x, byte y, byte c){
	byte pin_high;
	if ( x >= 8 ){
		//pin_high = ledMap[ y ];
		pin_high = pgm_read_byte(&ledMap[ y ]);
	} else {
		//pin_high = ledMap[ y + 8 ];
		pin_high = pgm_read_byte(&ledMap[ y + 8 ]);
	}
	if (c == 1) {
		if ( pin_high < 8 ){
		    workBuffer[ x*3     ] |= _BV( pin_high );
		}else if ( pin_high < 14 ){
		    workBuffer[ x*3 + 1 ] |= _BV( pin_high - 8 );
		}else{
		    workBuffer[ x*3 + 2 ] |= _BV( pin_high - 14 );
		}
	} else {
		if ( pin_high < 8 ){
		    workBuffer[ x*3     ] &= ~_BV( pin_high );
		}else if ( pin_high < 14 ){
		    workBuffer[ x*3 + 1 ] &= ~_BV( pin_high - 8 );
		}else{
		    workBuffer[ x*3 + 2 ] &= ~_BV( pin_high - 14 );
		}
	}
}

/* -----------------------------------------------------------------  */
byte ikkeiplexing::Get(byte x, byte y){
	byte pin_high;
	byte b;
	byte c;
	if ( x >= 8 ){
		//pin_high = ledMap[ y ];
		pin_high = pgm_read_byte(&ledMap[ y ]);
	} else {
		//pin_high = ledMap[ y + 8 ];
		pin_high = pgm_read_byte(&ledMap[ y + 8 ]);
	}

	if ( pin_high < 8 ){
	    b = workBuffer[ x*3     ] & _BV( pin_high );
	}else if ( pin_high < 14 ){
	    b = workBuffer[ x*3 + 1 ] & _BV( pin_high - 8 );
	}else{
	    b = workBuffer[ x*3 + 2 ] & _BV( pin_high - 14 );
	}

	return (b != 0);
}


/* -----------------------------------------------------------------  */
/** GetSW : get switch status
 */
byte ikkeiplexing::GetSW(byte edge){ // edge 0:level  1:positive edge 2:negative edge
	static byte previous_swdata;
	byte flag = 0;
	
	if (edge == 0){
		flag = swdata;
	} else if (edge == 1){
		flag = swdata & ~previous_swdata;
	} else if (edge == 2){
		flag = ~swdata & previous_swdata;
	}
	previous_swdata = swdata;
 	return (flag);
} 


/* -----------------------------------------------------------------  */
/** GetPeriod : get period flag
 */
byte ikkeiplexing::GetPeriod(void){
	static byte previous_period;
	byte flag = period & ~previous_period;
	previous_period = period;

 	return (flag);
} 


